try:
    import requests
except ModuleNotFoundError:
    requests = None

from . import isddc_time as itime


async def __init__(params: dict):
    return IsddcAPI(**params)


class IsddcAPI:
    token_timeout_point = 0

    def __init__(
            self,
            *,
            api: str,
            keystone: str,
            username: str,
            password: str,
            tenant_id: str = None,
            tenant_name: str = None,
            http_timeout: int = 60,
            token_timeout: int = 60 * 59,
    ):
        self.api = api
        self.keystone = keystone
        self.username = username
        self.password = password
        self.http_timeout = http_timeout
        self.token_timeout = token_timeout

        self.auth_api = f'http://{self.keystone}/v3/auth/tokens'
        self.scope = {'id': tenant_id} if tenant_id else {'name': tenant_name}

        self.headers = {
            'Content-Type': 'application/json',
            'X-Auth-Token': None
        }

    def __getattribute__(self, name):
        if name in ('get', 'post') and self.token_timeout_point < itime.Now.float:
            self.headers['X-Auth-Token']: str = requests.post(
                url=self.auth_api,
                json={'auth': {
                    'identity': {
                        'methods': ['password'],
                        'password': {'user': {
                            'domain': {'id': 'default'},
                            'name': self.username,
                            'password': self.password}}},
                    'scope': {'project': {
                        **self.scope,
                        'domain': {'id': 'default'}}}}}
            ).headers['X-Subject-Token']
            self.token_timeout_point = itime.Future(self.token_timeout).float

        return super().__getattribute__(name)

    def get(
            self,
            uri: str,
            *,
            params: dict = None,
            **kw
    ) -> requests.Response:
        res = requests.get(
            url=f'http://{self.api}{uri}',
            params=params,
            headers=self.headers,
            timeout=self.http_timeout,
            **kw)
        return self.check_status(res)

    def post(
            self,
            uri: str,
            *,
            data: dict = None,
            json: dict = None,
            **kw
    ) -> requests.Response:
        res = requests.post(
            url=f'http://{self.api}{uri}',
            data=data,
            json=json,
            headers=self.headers,
            timeout=self.http_timeout,
            **kw)
        return self.check_status(res)

    def report(
            self,
            data: dict,
            *,
            dtype: 'enum(metrics, inventories)'
    ) -> requests.Response:
        res = self.post(f'/v2.0/{dtype}', json=data)
        return self.check_status(res)

    @staticmethod
    def check_status(res: requests.Response) -> requests.Response:
        if res.status_code < 400:
            return res

        raise requests.HTTPError(
            f'api: %s, status: %s, info: %s'
            % (res.url, res.status_code, res.text))


def get(
        uri: str,
        *,
        params: dict = None,
        client: IsddcAPI = None,
        **kw
) -> requests.Response:
    return (client or __default__).get(
        uri, params=params, **kw)


def post(
        uri: str,
        *,
        data: dict = None,
        json: dict = None,
        client: IsddcAPI = None,
        **kw
) -> requests.Response:
    return (client or __default__).post(
        uri, data=data, json=json, **kw)


def report(
        data: dict,
        *,
        dtype: 'enum(metrics, inventories)',
        client: IsddcAPI = None
) -> requests.Response:
    return (client or __default__).report(
        data, dtype=dtype)


__default__: IsddcAPI
