try:
    from influxdb import InfluxDBClient
except ModuleNotFoundError:
    InfluxDBClient = None

from . import isddc_decorator as idec
from . import isddc_func      as ifunc

from .isddc_dict import IsddcDict as idict

__default__: InfluxDBClient


@idec.TryExceptAsync(Exception, name='InitInflux')
@idec.RetryAsync('InitInflux', count=10, cycle=10)
async def __init__(params: dict) -> InfluxDBClient:
    client = InfluxDBClient(**params)
    ifunc.OverFunc(client.close)
    return client


def query(
        sql: str,
        *,
        one: bool = False,
        epoch: str = None,
        dtype: 'enum(dict, list)' = dict,
        client: InfluxDBClient = None,
        **kw
) -> list or dict:
    """
    Encapsulate the 'InfluxDBClient.query' method.

    Help:
        Turn off the highlighting of the PyCharm SQL
        statement: `File -> Editor -> Inspections -> SQL`

    :param sql: SQL statement string.
    :param one: Returns a piece of data.
    :param epoch: The time format.
    :param dtype: The data type returned, dict or str
    :param client: Specified database, the default is `__default__`.
    """
    queryset = (client or __default__).query(sql, epoch=epoch, **kw)

    if not queryset:
        return {} if dtype is dict and one else []

    if dtype is dict:
        data = idict([[dict(
            (qst['columns'][i], val[i])
            for i in range(len(val)))
            for val in qst['values']]
            for qst in queryset.raw['series']])
    else:
        data = [series['values'] for series in queryset.raw['series']]

    if 'GROUP BY' not in sql:
        data = data[0]

    return data[0] if one else data


def write_points(
        data: list,
        *,
        timep: str = None,
        client: InfluxDBClient = None,
        **kw
) -> bool:
    return (client or __default__).write_points(
        points=data, time_precision=timep, **kw)
