import os
import time
import uuid
import asyncio
import threading as th
import multiprocessing as mp

from multiprocessing.pool import Pool
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import AsyncResult

from . import isddc_time as itime


class IsddcAsyncTasks:
    __tasks__ = {}

    def __new__(cls, *a, **kw):
        raise NotImplemented

    @classmethod
    def at(cls, async_func, *a, namespace: str = None, **kw):
        """Submit an asynchronous function and
        store it in the specified namespace."""
        if namespace not in cls.__tasks__:
            cls.__tasks__[namespace] = []
        cls.__tasks__[namespace].append((async_func, a, kw))

    @classmethod
    def batch_at(cls, async_func, *params) -> str:
        namespace: str = random_namespace()
        cls.__tasks__[namespace] = []

        for param in params:
            cls.at(async_func, param, namespace=namespace)

        return namespace

    @classmethod
    def run_until_complete(
            cls,
            namespace: str = None,
            *,
            timeout: int = None,
            return_when: str = asyncio.ALL_COMPLETED
    ) -> tuple:
        """Run all asynchronous function in the
        specified namespace until completed."""
        if not cls.check_namespace(namespace):
            raise RuntimeError(f'Namespace "{namespace}" not found.')

        loop = asyncio.new_event_loop()
        tasks = asyncio.wait(
            [fn(*a, **kw) for fn, a, kw in cls.__tasks__.pop(namespace)],
            loop=loop, timeout=timeout, return_when=return_when)

        try:
            return loop.run_until_complete(tasks)
        finally:
            loop.close()

    @classmethod
    def get(
            cls,
            namespace: str = None,
            *,
            timeout: int = None,
            return_when: str = asyncio.ALL_COMPLETED
    ) -> iter:
        """Run all asynchronous function in the specified
        namespace until complete, and get all task results."""
        done, pending = cls.run_until_complete(
            namespace, timeout=timeout, return_when=return_when)
        yield from (res.result() for res in done)

    @classmethod
    def check_namespace(cls, namespace: str) -> bool:
        return namespace in cls.__tasks__


def run_async_func(async_func, *a, **kw):
    """Run an asynchronous function and return its result."""
    loop = asyncio.get_event_loop()
    task = asyncio.ensure_future(async_func(*a, **kw), loop=loop)
    return loop.run_until_complete(task)


def random_namespace() -> str:
    """Return a random uuid4 string."""
    return uuid.uuid4().hex


class IsddcProcessPool(Pool):

    def __init__(
            self,
            name: str = None,
            *,
            processes: int = None,
            **kw
    ):
        name: str = name or type(self).__name__[:-4]
        processes: int = processes or (mp.cpu_count() or 1)

        super().__init__(
            processes=processes,
            initializer=self._initializer,
            initargs=[name],
            **kw)

        self.name = name
        self.processes = processes

        self.running = {}

    @staticmethod
    def _initializer(name: str):
        mp.current_process().name = f'{name}-{str(os.getpid())}'

    def at(
            self,
            func,
            *a,
            namespace: str = None,
            callback=None,
            error_callback=None,
            **kw
    ) -> AsyncResult:
        result: AsyncResult = self.apply_async(
            func=func, args=a, kwds=kw,
            callback=callback,
            error_callback=error_callback)

        if namespace not in self.running:
            self.running[namespace] = []
        self.running[namespace].append(result)

        return result

    def batch_at(
            self,
            func,
            *params,
            callback=None,
            error_callback=None
    ) -> random_namespace:
        namespace: str = random_namespace()
        self.running[namespace] = []

        for param in params:
            result: AsyncResult = self.at(
                func, param,
                namespace=namespace,
                callback=callback,
                error_callback=error_callback)
            self.running[namespace].append(result)

        return namespace

    def block_until_complete(
            self,
            namespace: str = None,
            *,
            delay_time: int = 0,
            timeout: int = None,
            ignore_timeout: bool = False,
            cycle: int or float = .1
    ) -> list:
        if not self.check_namespace(namespace):
            raise RuntimeError(f'Namespace "{namespace}"  not found.')

        time.sleep(delay_time)

        running: list = self.running.pop(namespace)
        results = []

        timeout_point: float = itime.Future(timeout or '36500d').float

        while running:
            time.sleep(cycle)

            for result in running.copy():
                if not result.ready():
                    continue
                if not result.successful():
                    raise result._value

                results.append(result._value)
                running.remove(result)

            if itime.Now.float >= timeout_point:
                if ignore_timeout:
                    break
                raise TimeoutError(
                    f'Waiting for result timeout, time out {len(running)} results.')

        return results

    def get(self):
        raise NotImplemented

    def batch_at_and_wait_complete(
            self,
            func,
            *params,
            **kw
    ) -> list:
        namespace: str = self.batch_at(
            func, *params,
            callback=kw.pop('callback', None),
            error_callback=kw.pop('error_callback', None))
        return self.block_until_complete(namespace, **kw)

    def check_namespace(self, namespace: str) -> bool:
        return namespace in self.running

    @staticmethod
    def block(sleep: int):
        time.sleep(sleep)


class IsddcThreadPool(IsddcProcessPool, ThreadPool):

    @staticmethod
    def _initializer(name: str):
        th.current_thread().name = f'{name}'
