import os
import logging
import asyncio

from types import *
from typing import *

from multiprocessing.pool import Pool
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import AsyncResult

try:
    from requests import Response as HttpResponse
    from requests import HTTPError
except ModuleNotFoundError:
    HttpResponse = HTTPError = None
try:
    from dbutils.pooled_db import PooledDB
except ModuleNotFoundError:
    PooledDB = None
try:
    from influxdb import InfluxDBClient
except ModuleNotFoundError:
    InfluxDBClient = None
try:
    from pykafka import Topic as KafkaTopic
except ModuleNotFoundError:
    KafkaTopic = None
try:
    from pymongo import MongoClient
    from pymongo.database import Database as MongoDatabase
    from pymongo.collection import Collection as MongoCollection
except ModuleNotFoundError:
    MongoClient = MongoDatabase = MongoCollection = None
try:
    from pymysql.connections import Connection as MySQLConnection
except ModuleNotFoundError:
    MySQLConnection = None
try:
    from paramiko import SSHClient
except ModuleNotFoundError:
    SSHClient = None

from . import isddc_baseclass as ibase
from . import isddc_func      as ifunc
from . import isddc_mysql     as imysql


class IsddcAliyun(metaclass=ibase.ObjectMode):
    """isddc_aliyun"""

    init_params_sample = {
        'init': {
            'timeout': 60,
            'method': 'post',
            'protocol_type': 'https',
            'accept_format': 'json',
        },
        'send_sms': {
            'name': 'SendSms',
            'ak': None,
            'secret': None,
            'region_id': 'cn-hangzhou',
            'domain': 'dysmsapi.aliyuncs.com',
            'version': '2017-05-25',
            'SignName': '歌琪怡玲萍云',
            'TemplateCode': 'SMS_199195663',
            'PhoneNumbers': 1388888888
        },
        'send_mail': {
            'name': 'SingleSendMail',
            'ak': None,
            'secret': None,
            'region_id': 'cn-hangzhou',
            'domain': 'dm.aliyuncs.com',
            'version': '2015-11-23',
            'AddressType': 1,
            'ReplyToAddress': True,
            'AccountName': 'trigger@mail.gqylpy.com',
            'ToAddress': 'zhuyk4@lenovo.com',
            'FromAlias': 'gqylpy-base'
        }
    }

    class Aliyun:
        def __init__(self, action: dict):
            pass

    @staticmethod
    def send_sms(**active_params):
        pass

    @staticmethod
    def send_mail(**active_params):
        pass


class IsddcAPI(metaclass=ibase.ObjectMode):
    """isddc_api"""

    init_params_sample = {
        'cmo109': {
            'api': '10.121.134.109:30870',
            'keystone': '10.121.134.109:31500',
            'username': None,
            'password': None,
            'tenant_id': None,
            'http_timeout': 60,
            'token_timeout': 60 * 59
        },
        'oi50': {
            'api': '10.121.218.44:30870',
            'keystone': '10.121.218.50:5000',
            'username': 'admin',
            'password': 'engine',
            'tenant_id': 'admin',
            'http_timeout': 60,
            'token_timeout': 60 * 10
        }
    }

    class IsddcAPI:
        """Born for OI isddc-api."""

        def __init__(
                self,
                *,
                api: str,
                keystone: str,
                username: str,
                password: str,
                tenant_id: str = None,
                tenant_name: str = None,
                http_timeout: int = 60,
                token_timeout: int = 60 * 59,
        ):
            """
            @param api:           address:port of isddc-api service.
            @param keystone:      address:port of keystone service.
            @param username:      Keystone auth username.
            @param password:      Keystone auth password.
            @param tenant_id:     Keystone auth tenant id, or @param(tenant_name), high priority.
            @param tenant_name:   Keystone auth tenant name, or @param(tenant_id).
            @param http_timeout:  Http timeout.
            @param token_timeout: Token timeout.
            """

        def get(
                self,
                uri: str,
                *,
                params: dict = None,
                **kw
        ) -> HttpResponse:
            """"Get data from OI isddc-api."""

        def post(
                self,
                uri: str,
                *,
                data: dict = None,
                json: dict = None,
                **kw
        ) -> HttpResponse:
            """Post data to OI isddc-api."""

        def report(
                self,
                data: dict,
                *,
                dtype: 'enum(metrics, inventories)'
        ) -> HttpResponse:
            """Report data, metrics or inventories."""

        @staticmethod
        def check_status(res: HttpResponse) -> HttpResponse:
            if res.status_code < 400:
                return res
            raise HTTPError

    @staticmethod
    def get(
            uri: str,
            *,
            params: dict = None,
            client: IsddcAPI = None,
            **kw
    ) -> HttpResponse:
        """Get data from OI isddc-api."""

    @staticmethod
    def post(
            uri: str,
            *,
            data: dict = None,
            json: dict = None,
            client: IsddcAPI = None,
            **kw
    ) -> HttpResponse:
        """Post data to OI isddc-api."""

    @staticmethod
    def report(
            data: dict,
            *,
            dtype: 'enum(metrics, inventories)',
            client: IsddcAPI = None
    ) -> HttpResponse:
        """Report data, metrics or inventories."""


class IsddcBaseClass(metaclass=ibase.ObjectMode):
    """isddc_baseclass"""

    class SingletonMode:
        """Inherit me, you will
        become a singleton class."""

    class DictMode:
        """Inherit me, You will have
        the grammar of the 'dict'."""

        def __getitem__(self, name):
            pass

        def __setitem__(self, name, value):
            pass

        def __delitem__(self, name):
            pass

    class IterMode:
        """Inherit me, your object can be iterated,
        and can be called by the 'if in' statement."""

        def __iter__(self):
            pass

    class ObjectMode(type):
        """To essentially turn a class into an object."""

    class FuncMode(type):
        """Enable classes to be called as functions.
        Instantiating class no longer get the object, it
        directly gets the return value of its __call__ method."""


class IsddcCommand(metaclass=ibase.ObjectMode):
    """isddc_command"""

    class IsddcCommand:
        """
        Executes the incoming command directly at instantiation time, and saves:
            self.code       = Command exit code
            self.raw_output = Command output
        """
        code: int
        raw_output: str

        def __init__(self, cmd: str, *, timeout: int = None):
            """
            @param cmd:     Command string.
            @param timeout: Command timeout, Default never times out.
            """

        def raise_if_error(self):
            if self.code != 0:
                raise self.CommandError

        @property
        def status(self) -> bool:
            if self.code == 0:
                return True
            return False

        @property
        def output(self) -> str:
            return process(self.raw_output)

        @property
        def status_output(self) -> tuple:
            return self.status, self.output

        @property
        def code_output(self) -> tuple:
            return self.code, self.output

        @property
        def output_else_raise(self) -> str:
            if self.code == 0:
                return self.output
            raise self.CommandError

        @property
        def output_else_define(self, define: Any = None):
            if self.code == 0:
                return self.output
            return define

        def is_contain_string(self, string: str) -> bool:
            if string in self.raw_output:
                return True
            return False

        def output_if_contain_string_else_raise(self, string: str) -> str:
            if string in self.raw_output:
                return self.output
            raise self.CommandError

        def table_output_to_dict(self, **kw) -> list:
            """Can turn command results printed in tabular format
            into dictionaries, such as `kubectl get pod -n isddc`."""

        @classmethod
        def table2dict(cls, table: str, *, split: str = None) -> list:
            """Can turn command results printed in tabular format
            into dictionaries, such as `kubectl get pod -n isddc`."""

        class CommandError(Exception):
            pass


class IsddcCrypto(metaclass=ibase.ObjectMode):
    """isddc_crypto"""

    class IsddcCrypto:
        """A reversible encryption technique based on Unicode Code."""

        def __init__(self, offset: int = -1):
            """
            @param offset:
                Specifies the Unicode Code offset.
                Encryption and decryption must use the same value.
            """

        def encrypt(self, cleartext: str) -> str:
            """
            @param cleartext: The string to be encrypt.
            @return:          The encrypted value.
            """

        def decrypt(self, ciphertext: str) -> str:
            """
            @param ciphertext: The string to be decrypt.
            @return:           The decrypted value.
            """

    class MD5:

        def __new__(cls, data: Any = None):
            """
            @param data: Data to be encrypt.

            if data is None:
                return object.__new__(cls)
            else:
                return md5-value of data.
            """

        def update(self, data: Any):
            """Add or append a segment of data to encrypt."""

        def update_str(self, data: str):
            """Add or append a string to encrypt."""

        @property
        def hexdigest(self) -> str:
            """Execute encryption program and return md5-value."""

    def encrypt(self, cleartext: str, *, offset: int = -1) -> str:
        """Encrypts the string and returns the encrypted value."""
        return self.IsddcCrypto(offset).encrypt(cleartext)

    def decrypt(self, ciphertext: str, *, offset: int = -1) -> str:
        """Decrypt the ciphertext and return the decrypted value."""
        return self.IsddcCrypto(offset).decrypt(ciphertext)


class IsddcDBPool(metaclass=ibase.ObjectMode):
    """isddc_dbpool"""

    init_params_sample = {
        'init': {
            'mincached': 0,
            'maxcached': 1,
            'maxshared': 0,
            'maxconnections': 2,
            'blocking': True,
            'maxusage': None,
            'setsession': None,
            'ping': 1
        },
        'db1': {
            'creator': 'pymysql',
            'db': 'isddc',
            'host': 'localhost',
            'port': 3306,
            'user': 'isddc',
            'passwd': 'user@isddc',
            'charset': 'utf8',
            'autocommit': False,
            'use_unicode': None,
            'read_timeout': None,
            'write_timeout': None,
            'connect_timeout': None,
        },
        'db2': {...}
    }

    class mysql(imysql.Transaction):

        def __new__(
                cls,
                sql: str = None,
                *,
                one: bool = False,
                dtype: Union[list, dict] = dict,
                pool: PooledDB = None
        ):
            """
            if sql is not None:
                Execute the sql and return the results.
            else:
                Continue calling __init__ and get an instance of mysql.
            """
            pass

        def __init__(
                self,
                *,
                dtype: Union[list, dict] = dict,
                pool: PooledDB = None
        ):
            pass

        def __enter__(self):
            pass

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass


class IsddcDecorator(metaclass=ibase.ObjectMode):
    """isddc_decorator"""

    class TryExcept:
        """Exception handler.
        The function that is decorated will have exception handling capabilities.

        #-- Simple demo:
            @TryExcept(ValueError)
            def func():
                int('a')
        """

        def __init__(
                self,
                exc_type: Union[type, tuple],
                *,
                name: str = None,
                ignore: bool = False,
                exc_return: Any = None,
                exc_callback: Callable = None
        ):
            """
            @param exc_type:     Which exceptions to handle.
            @param name:         Exception identification, which is displayed in the exception output.
            @param ignore:       Whether to ignore an exception or not.
                                 If so, no exception information will be logged.
            @param exc_return:   The value returned after the exception was raised.
            @param exc_callback: The function that is called after the exception is raised.
            """

        def __call__(self):
            pass

    class Retry:
        """
        When an exception is raised, attempt to reexecute.

        #-- Simple demo:
            @Retry(count=3, cycle=1)
            def func():
                int('a')

        #-- Use in conjunction with TryExcept will support
            both exception retry and exception handling:
            @TryExcept(ValueError)
            @Retry(count=3, cycle=1)
            def func():
                int('a')
        """

        def __init__(
                self,
                name: str = None,
                *,
                count: int = None,
                cycle: int = 0,
                retry_exc: Union[type, tuple] = Exception,
        ):
            """
            @param name:      Exception identification which is displayed in the exception output.
            @param count:     Total number of executions.
            @param cycle:     Seconds between each attempt.
            @param retry_exc: Which exceptions are supported to retry.
            """

        def __call__(self):
            pass

    class WhileTrue:
        """
        The decorated function will be executed in an infinite
        loop, Can end the loop by returning a string '_break_'.

        #-- Simple demo:
            @WhileTrue(cycle=1, minus_run_time=True)
            def func():
                print(itime.Now.str)

        #-- To exit the loop, use syntax `return '_break_'`:
            @WhileTrue(cycle=1, minus_run_time=True)
            def func():
                print(itime.Now.str)
                return '_break_'

        #-- Use in conjunction with TryExcept will exceptions
            that are raised in each cycle can be handled:
            @WhileTrue(cycle=1, minus_run_time=True)
            @TryExcept(Exception)
            def func():
                print(itime.Now.str)
                int('a')
        """

        def __init__(
                self,
                cycle: Union[int, float] = 0,
                *,
                minus_run_time: bool = False,
                wait_first: bool = False
        ):
            """
            @param cycle:          Loop interval.
            @param minus_run_time: Loop interval minus function run time.
                                   If True, the cycle period will be more accurate.
            @param wait_first:     If True, the first execution will wait for @param(cycle) seconds.
            """

        def __call__(self):
            pass

    class RunTime:
        """
        Record the execution time of the function.

        #-- Simple demo:
            @idec.RunTime()
            def func():
                time.sleep(1.234)
        """

        def __init__(self, name: str = None):
            """
            @param name: Exception identification which is displayed in the exception output.
            """

        def __call__(self):
            pass

    class TestFuncSpeed:

        def __init__(self, count: int, *, keep: int = 7):
            pass

        def __call__(self):
            pass

    class BeforeFunc:
        """
        Execute another function before another function.

        #-- Simple demo:
            def before_func():
                print('exec before_func')

            @BeforeFunc(before_func)
            def func():
                print('exec func')

        #-- When have parameters:
            def before_func(param1, param2):
                print('exec before_func', param1, param2)

            @BeforeFunc(before_func)
            def func(param1, param2):
                print('exec func', param1, param2)

        #-- Modify parameters in 'before_func':
            def before_func(param1, param2):
                print('exec before_func', param1, param2)
                return {'param1': param1.upper()}

            @BeforeFunc(before_func)
            def func(param1, param2):
                print('exec func', param1, param2)
        """

        def __init__(self, before_func: Callable, *, enable: bool = True):
            """
            @param before_func: A func.
            @param enable:      If False, then @param(before_func) is not called.
            """

        def __call__(self):
            pass

    class AfterFunc:
        """
        Executes another function after another function.

        #-- Simple demo:
            def after_func():
                print('exec after_func')

            @idec.AfterFunc(after_func)
            def func():
                print('exec func')

        #-- When have parameters:
            def after_func(param1, param2):
                print('exec after_func', param1, param2)

            @idec.AfterFunc(after_func)
            def func(param1, param2):
                print('exec func', param1, param2)
                return param1.upper(), param2.upper()

        #-- @param(use_raw_params) usage:
            def after_func(param1, param2):
                print('exec after_func', param1, param2)

            @idec.AfterFunc(after_func, use_raw_params=True)
            def func(param1, param2):
                print('exec func', param1, param2)
        """

        def __init__(
                self,
                after_func: Callable,
                *,
                use_raw_params: bool = False,
                enable: bool = True
        ):
            """
            @param after_func:     A func.
            @param use_raw_params: See the above usage.
            @param enable:         If False, then @param(after_func) is not called.
            """

        def __call__(self):
            pass

    class TryExceptAsync(TryExcept):
        """Exception handler, used for asynchronous
        functions, Usage is the same as TryExcept."""

        def __call__(self):
            pass

    class RetryAsync(Retry):
        """When an exception is raised, attempt to reexecute, used
        for asynchronous functions, Usage is the same as Retry."""

        def __call__(self):
            pass


class IsddcDict(metaclass=ibase.ObjectMode):
    """isddc_dict"""

    class IsddcDict(dict):
        """IsddcDict == dict
        I inherited 'dict', my main function is to
        allow the dict to get or set values by `d.key`.
        """

        def __init__(self, *a, **kw):
            pass

        def __getattribute__(self, name):
            pass


class IsddcFileSystem(metaclass=ibase.ObjectMode):
    """isddc_filesystem"""

    class FileDataOperator(ibase.DictMode):
        """
        db = ifs.FileDataOperator('db/')

        # Write content to file 'db/test.txt'
        db['test.txt'] = ...

        # Read content from file 'db/test.txt'
        text = db['test.txt']

        # Delete file 'db/test.txt'
        del db['test.txt']

        # Deep directory:
        db['dir1/dir2/test.txt'] = ...
        text = db['dir1/dir2/test.txt']
        """

        def __init__(self, db_dir: str):
            """
            @param db_dir:
                Specify a directory path, it could be a relative path.
                If the specified directory does not exist, it is created.
            """

        def __iter__(self):
            pass

    class Filetor:
        """Writes data to or reads data from a file based on its type."""

        def __new__(
                cls,
                file: str,
                data: Any = None,
                *,
                ftype: 'enum(json, yaml, obj, txt)' = None,
                encoding: str = 'UTF-8',
                **kw
        ):
            """
            @param file:     A file path.
            @param data:     The data to be written to the file, if None,
                             the data is read from the file and return.
            @param ftype:    Specifies File type, otherwise is determined by file suffix name.
            @param encoding: Open the encoding of the file.

            if @param(data) is None:
                Read contents of @param(file) according to @param(ftype) and return.
            else:
                Write @param(data) to @param(file) according to @param(ftype).
            """

    @staticmethod
    def fetch_deep_path(path: str = None) -> IsddcDict.IsddcDict:
        """Get all file and directory paths under @param(root)."""

    @staticmethod
    def joinpath2abs(*a) -> str:
        """Return its absolute path after concatenating the path."""

    @staticmethod
    def dirname(path: str, *, level: int = 1) -> str:
        """Culling directories from the
        end according to `level`, usually
        used to get the project root path.
        """

    @staticmethod
    def gendir(*a) -> str:
        """Generate path
        Generates a directory based on the passed in value,
        creates it if it does not exist, and returns it.
        """

    isabs = os.path.isabs
    isdir = os.path.isdir
    isfile = os.path.isfile
    exists = os.path.exists
    joinpath = os.path.join
    abspath = os.path.abspath
    basename = os.path.basename


class IsddcInflux(metaclass=ibase.ObjectMode):
    """isddc_influx"""

    init_params_sample = {
        'db1': {
            'host': 'localhost',
            'port': 8086,
            'database': 'isddc'
        },
        'db2': {...}
    }

    @staticmethod
    def query(
            sql: str,
            *,
            one: bool = False,
            epoch: str = None,
            dtype: Union[dict, list] = dict,
            client: InfluxDBClient = None,
            **kw
    ) -> Union[dict, list]:
        """
        Encapsulate the influxdb.InfluxDBClient.query method.

        @param sql:    SQL statement string.
        @param one:    Returns a piece of data.
        @param epoch:  The time format.
        @param dtype:  The data type returned, dict or str
        @param client: Specified database, the first is used by default.
        """

    @staticmethod
    def write_points(
            data: list,
            *,
            timep: str = None,
            client: InfluxDBClient = None,
            **kw
    ) -> bool:
        """Encapsulate the influxdb.InfluxDBClient.write_points method."""


class IsddcKafka(metaclass=ibase.ObjectMode):
    """isddc_kafka"""

    init_params_sample = {
        'queue1': {
            'hosts': 'localhost:31092',
            'topic': 'isddc'
        },
        'conn1': {
            'hosts': 'localhost:31092',
            'topic': None
        }
    }

    @staticmethod
    def send(msg: Any, *, topic: KafkaTopic = None):
        """Simple Kafka message sender."""


class IsddcLog(metaclass=ibase.ObjectMode):
    """isddc_log"""

    init_params_sample = {
        'simple': {
            'level': 'NOTSET',
            'output': 'stream',
            'datefmt': '%F %T',
            'logfmt': '[%(asctime)s] [%(levelname)s] %(message)s'
        },
        'sine': {
            'level': 'INFO',
            'output': 'stream, file',
            'handler': '/var/log/isddc.log',
            'datefmt': '%Y-%m-%d %H:%M:%S',
            'logfmt': '[%(asctime)s] [%(module)s.%(funcName)s line:%(lineno)d] [%(levelname)s] %(message)s'
        },
        'logger3': {...}
    }

    @staticmethod
    def debug(msg: str, *, logger: logging.Logger = None):
        pass

    @staticmethod
    def info(msg: str, *, logger: logging.Logger = None):
        pass

    @staticmethod
    def warning(msg: str, *, logger: logging.Logger = None):
        pass

    @staticmethod
    def error(msg: str, *, logger: logging.Logger = None):
        pass

    @staticmethod
    def critical(msg: str, *, logger: logging.Logger = None):
        pass

    simple: logging.Logger = None
    sine: logging.Logger = None


class IsddcKube(metaclass=ibase.ObjectMode):
    """isddc_kube"""

    init_params_sample = {
        'init': {
            'protocol': 'https',
            'verify_ssl': False,
            'load_api': 'CoreV1Api, AppsV1Api, ...'
        },
        'isddc': {
            'api_server': 'localhost:6443',
            'token': None
        },
        'oi49': {...}
    }


class IsddcMongo(metaclass=ibase.ObjectMode):
    """isddc_mongo"""

    init_params_sample = {
        'db1': {
            'host': 'localhost',
            'port': 27017,
            'db': 'isddc'
        },
        'db2': {...}
    }

    @staticmethod
    def specify_collection(
            collection: str,
            *,
            db: Union[MongoDatabase, MongoClient] = None
    ) -> MongoCollection:
        """Specify a collection and return the collection object."""


class IsddcMultiTask(metaclass=ibase.ObjectMode):
    """isddc_multitask"""

    class IsddcAsyncTasks:

        def __new__(cls, *a, **kw):
            raise NotImplemented

        @classmethod
        def at(
                cls,
                async_func: Callable,
                *a,
                namespace: str = None,
                **kw
        ):
            """Submit an asynchronous function and
            store it in the specified namespace."""

        @classmethod
        def batch_at(
                cls,
                async_func: Callable,
                *params
        ) -> str:
            """Batch submit async task."""

        @classmethod
        def run_until_complete(
                cls,
                namespace: str = None,
                *,
                timeout: int = None,
                return_when: str = asyncio.ALL_COMPLETED
        ) -> tuple:
            """Run all asynchronous function in the
            specified namespace until completed."""

        @classmethod
        def get(
                cls,
                namespace: str = None,
                *,
                timeout: int = None,
                return_when: str = asyncio.ALL_COMPLETED
        ) -> Iterator:
            """Run all asynchronous function in the specified
            namespace until complete, and get all task results."""

        @classmethod
        def check_namespace(cls, namespace: str) -> bool:
            """Check if the namespace exists."""

    @staticmethod
    def run_async_func(async_func: Callable, *a, **kw) -> Any:
        """Run an asynchronous function and return its result."""

    @staticmethod
    def random_namespace() -> str:
        """Return a random uuid4 string."""

    class IsddcProcessPool(Pool):

        def __init__(
                self,
                name: str = None,
                *,
                processes: int = None,
                **kw
        ):
            """
            Create a Process pool.
            @param name:      Pool name.
            @param processes: Process number.
            """

        def at(
                self,
                func: Callable,
                *a,
                namespace: str = None,
                callback: Callable = None,
                error_callback: Callable = None,
                **kw
        ) -> AsyncResult:
            """
            Submit a task to the pool.
            @param func:           A function.
            @param a:              Position parameters of @param(func).
            @param namespace:      Specifies the  name space of @param(func)
                                   and creates it if it does not exist.
            @param callback:       Callback function of @param(func).
            @param error_callback: Error callback function of @param(func).
            @param kw:             Keyword parameters of @param(func).
            @return:               Result object.
            """

        def batch_at(
                self,
                func: Callable,
                *params,
                callback: Callable = None,
                error_callback: Callable = None
        ) -> str:
            """Batch submit tasks to the pool, and returns
            the namespace in which these tasks reside."""

        def block_until_complete(
                self,
                namespace: str = None,
                *,
                delay_time: int = 0,
                timeout: int = None,
                ignore_timeout: bool = False,
                cycle: Union[int, float] = .1
        ) -> list:
            """
            Wait for all tasks in the specified namespace to complete.
            @param namespace:      A namespace.
            @param delay_time:     How many seconds to wait before entering the loop check phase.
            @param timeout:        Tasks timeouts.
            @param ignore_timeout: If true, a task timeout will throw an exception.
            @param cycle:          The loop checks the duration of the task.
            @return:               Tasks result.
            """

        def get(self):
            raise NotImplemented

        def batch_at_and_wait_complete(
                self,
                func: Callable,
                *params,
                **kw
        ) -> list:
            pass

        def check_namespace(self, namespace: str) -> bool:
            pass

        @staticmethod
        def block(sleep: int):
            """The actual execution will be time.sleep()."""
            pass

    class IsddcThreadPool(IsddcProcessPool, ThreadPool):
        def __init__(
                self,
                name: str = None,
                *,
                processes: int = None,
                **kw
        ):
            """
            Create a thread pool.
            @param name:      Pool name.
            @param processes: Thread number.
            """


class IsddcMysql(metaclass=ibase.ObjectMode):
    """isddc_mysql"""

    init_params_sample = {
        'init': {
            'charset': 'utf8',
            'autocommit': False,
            'use_unicode': None,
            'read_timeout': None,
            'write_timeout': None,
            'connect_timeout': None
        },
        'db1': {
            'db': 'isddc',
            'host': 'localhost',
            'port': 3306,
            'user': 'isddc',
            'passwd': 'user@isddc',
            'charset': 'utf8'
        },
        'db2': {...}
    }

    @staticmethod
    def sql(
            s: str,
            *,
            one: bool = False,
            dtype: Union[dict, tuple] = dict,
            conn: MySQLConnection = None
    ):
        """Execute an SQL statement.
        @param s:     SQL statement.
        @param one:   If true, the first data in the query set will be
                      returned, which is suitable for single data query.
        @param dtype: Data type, dictionary or tuple.
        @param conn:  Database connection, First initialized connection is used by default.
        @return:      Queryset or commit result.
        """

    class Transaction:
        """Execute MySQL transactions."""

        def __init__(
                self,
                dtype: Union[dict, tuple] = dict,
                *,
                conn: MySQLConnection = None
        ):
            """
            @param dtype: Data type, dictionary or tuple.
            @param conn:  Database connection, First initialized connection is used by default.
            """

        def __call__(self, s: str) -> int:
            pass

        def __enter__(self):
            pass

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

        def __invert__(self):
            pass

        @staticmethod
        def upgrade_dtype(data: list = None) -> list:
            pass

        @property
        def fetchone(self) -> Union[tuple, dict]:
            pass

        @property
        def fetchall(self) -> Union[tuple, dict]:
            pass

        @property
        def last_insert_id(self) -> int:
            """Returns the id of the data that
            was last inserted by `self.conn`."""


class IsddcSSH(metaclass=ibase.ObjectMode):
    """isddc_ssh"""

    init_params_sample = {
        'server1': {
            'hostname': '10.121.12.58',
            'port': 22,
            'username': 'root',
            'password': 'user@root'
        },
        'server2': {...}
    }

    @staticmethod
    def cmd(
            command: str,
            *,
            client: SSHClient = None
    ) -> str:
        """Executes @param(command) and
        returns the result of the command."""

    @staticmethod
    def put_file(
            local: str = None,
            remote: str = None,
            *,
            client: SSHClient = None
    ):
        """
        Sends a file to the remote host.
        @param local:  Local file path.
        @param remote: Remote file path.
        """

    @staticmethod
    def get_file(
            remote: str = None,
            local: str = None,
            *,
            client: SSHClient = None
    ):
        """
        Pull a file from the remote host.
        @param remote: Remote file path.
        @param local: Local file path.
        """


class IsddcTime(metaclass=ibase.ObjectMode):
    """isddc_time"""

    class Now(metaclass=ibase.ObjectMode):
        """
        Get a current time, usage:
            now_int: int = itime.Now.int
            now_str: str = itime.Now.str
        """

        @property
        def int(self) -> int:
            pass

        @property
        def float(self) -> float:
            pass

        @property
        def struct(self) -> tuple:
            pass

        @property
        def str(self) -> str:
            pass

        @staticmethod
        def strf(format: str) -> str:
            """
            @param format: Time format.
            @return:       String format time.
            """

    class Future:
        """
        Get a point in time in the future, usage:
            future_time_point_1hour_int: int = itime.Future('1h').int
            future_time_point_1minute_str: str = itime.Future(60).str
        """

        def __init__(self, future_time: Union[int, float, str]):
            """
            @param future_time: How long after.
            """

        @property
        def int(self) -> int:
            pass

        @property
        def float(self) -> float:
            pass

        @property
        def struct(self) -> tuple:
            pass

        @property
        def str(self) -> str:
            pass

        def strf(self, format: str) -> str:
            """
            @param format: Time format.
            @return:       String format time.
            """

    class Past(Future):
        """Get a point in time in the past, Use with Future."""

        def __init__(self, past_time: Union[int, float, str]):
            """
            @param past_time: How long before.
            """

    @staticmethod
    def stamp2str(
            stamp: Union[int, float] = None,
            format: str = '%F %T',
            *,
            gmt: bool = False
    ) -> str:
        """timestamp to string time"""

    @staticmethod
    def str2stamp(
            string: str,
            format: str = '%Y-%m-%d %H:%M:%S'
    ) -> float:
        """string time to timestamp"""

    class Time2Second:
        """
        Convert a unit time to a second.

        Usage:
            second_1d: int = itime.Time2Second('1d')
            second_1h30m: int = itime.Time2Second('1h30m')

        Contrary to the 'Second2Time' function below.
        """

        def __new__(cls, unit_time: str) -> Union[int, float]:
            """
            @param unit_time: Can be: 1d, 1h, 1m, 1s, 1h30m, 1.5h
            @return:          Number of seconds.
            """

    class Second2Time:
        """Pass in a time in seconds, and
        it will be converted to unit time."""

        def __new__(cls, sec: Union[int, float]) -> str:
            """
            @param sec: A Number of seconds.
            @return:    A unit time of string.
            """


class IsddcFunc(metaclass=ibase.ObjectMode):
    """isddc_func"""

    linux: bool = None
    container: bool = None

    class DictInterProcess(metaclass=ibase.FuncMode):
        """Dictionary internal processing
        Used to batch process values inside a dictionary.

        #-- Simple demo:
            data = {'a': 'a', 'b': 'b'}

            def func(name, value):
                if name == 'b':
                    return value.upper()

            func(data, func)
            print(data)
        """

        def __init__(
                self,
                data: dict,
                func: Callable = None
        ):
            """@param(func) process @param(data)."""

    class BootParams:
        """Get boot parameters.
        Boot parameters with default values
        are available through **default_params.
        """

        def __new__(cls, *params, **default_params) -> dict:
            """
            @param params:         Gets which startup parameters.
            @param default_params: Startup parameters with default values.
            """

    class Import:
        """Import a module, variable, or object."""

        @classmethod
        def module(cls, module: str):
            """
            @param module: A module path of string, e.g. isddc.isddc_func
            @return:       A object of @param(module).
            """

        @classmethod
        def name(cls, name: str):
            """
            @param name: A variable path of string, e.g. isddc.isddc_func.IsddcTools
            @return:     A variable.
            """

        @classmethod
        def object(cls, class_path: str, *a, **kw):
            """
            @param class_path: A class path of string, e.g. isddc.isddc_func.IsddcTools
            @param a:          Class instantiates parameters.
            @param kw:         Class instantiates parameters.
            @return:           A object of @param(class_path).
            """

        @classmethod
        def try_import(cls, module: str, default: Any = None):
            """
            @param module:  A module path of string, e.g. isddc.isddc_func
            @param default: Default return value
            @return:        A object of @param(module), or @param(default)
            """

    class OverFunc:
        """Functions passed in during instantiation
        are executed at after the main program is finished."""

        def __init__(self, func: Callable, *a, **kw):
            """
            Add a function that will execute after the main program is finished.
            @param func: A function to be executed at after the main program is finished.
            @param a:    Function parameters.
            @param kw:   Function parameters.
            """

    @staticmethod
    def underline2hump(name: str) -> str:
        """Change the underscore name to the hump name."""

    @staticmethod
    def hump2underline(name: str) -> str:
        """Change the hump name to the underscore name."""

    @staticmethod
    def get_caller(level: int = 1) -> str:
        """Gets the location where the current function was called."""

    @staticmethod
    def prt(
            *args,
            color: int = 31,
            font: int = 0,
            end: str = '\n',
            sep: str = ' ',
            file: open = None
    ):
        """
        Custom print font and color.

        Color:
            31-red 32-green 34-blue 37-Gray 38-black
        Font:
            0-normal 1-bold
        sep:
            string inserted between values, default a space.
        """

    @staticmethod
    def avg(numbers: list) -> float:
        """Return average value of @param(numbers)."""

    @staticmethod
    def count_occurrences(data: list, value: Any) -> int:
        """Return number of occurrences of @param(value) in @param(data)."""

    @staticmethod
    def max_value_index(values: list) -> int:
        """Return the index of the maximum value in @param(values)."""


class DenyImport(metaclass=ibase.ObjectMode):

    def __getattribute__(self, name):
        raise ImportError('The import method that is not recommended.')


module_mapping = {}

for iname, iobj in globals().copy().items():
    if iname[:5] == 'Isddc':
        for attr in dir(iobj):
            if attr[0] != '_':
                module_str = f'{__name__[:-13]}.{iobj.__doc__}'
                imodule: ModuleType = ifunc.Import.module(module_str)
                module_mapping[imodule] = iobj
                setattr(iobj, attr, getattr(imodule, attr, None))

IsddcInit = ifunc.__init__
IsddcCommand = IsddcCommand.IsddcCommand
IsddcDict = IsddcDict.IsddcDict
IsddcAsync = IsddcMultiTask.IsddcAsyncTasks
